"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.LinkedTree = exports.LinkedTreeNode = void 0;
const tslib_1 = require("tslib");
const omit_1 = tslib_1.__importDefault(require("lodash/omit"));
const object_1 = require("../../common/object");
const types_1 = require("../../common/types");
const utils_1 = require("../tree/utils");
const utils_2 = require("./utils");
class LinkedTreeNode {
    constructor(data, clone = false) {
        this._children = [];
        this.data = clone ? object_1.deepClone(data) : data;
    }
    get parent() { return this._parent; }
    get prev() { return this._prev; }
    get next() { return this._next; }
    get firstChild() { return this._firstChild; }
    get lastChild() { return this._lastChild; }
    get children() { return this._children; }
    get isLeaf() { return this._children.length === 0; }
    get hasParent() { return this._parent !== undefined; }
    hasAncestor(ancestor) {
        let has = false;
        utils_2.bubble(this, (current) => {
            has = current === ancestor;
            return !has;
        });
        return has;
    }
    getNodeHeight(root) {
        let distance = 0;
        if (!this.hasParent)
            return distance;
        if (root && this === root)
            return distance;
        let complete = utils_2.bubble(this, (current) => {
            distance++;
            if (root) {
                return current !== root;
            }
        });
        return root && complete ? -1 : distance;
    }
}
exports.LinkedTreeNode = LinkedTreeNode;
class LinkedTree {
    constructor(data = {}) {
        this._root = new LinkedTreeNode(data);
    }
    get root() { return this._root; }
    isRoot(target) {
        this._root === target;
    }
    setRoot(node) {
        this._root = node instanceof LinkedTreeNode
            ? node
            : new LinkedTreeNode(node);
        return this;
    }
    //#region add node 
    addToBefore(source, target) {
        if (!target.hasParent) {
            utils_2.insert(source, target.parent);
            utils_2.setBefore(source, target);
        }
        return this;
    }
    addToAfter(source, target) {
        if (!target.hasParent) {
            utils_2.insert(source, target.parent);
            utils_2.setAfter(source, target);
        }
        return this;
    }
    addToChildrenFirst(source, target) {
        utils_2.insert(source, target);
        utils_2.setFirst(source, target);
        return this;
    }
    addToChildrenLast(source, target) {
        utils_2.insert(source, target);
        utils_2.setLast(source, target);
        return this;
    }
    //#endregion
    //#region move node 
    _canMoveRemove(source) {
        if (source.hasParent) {
            utils_2.remove(source);
            return true;
        }
        return false;
    }
    _canMoveToSibling(source, target) {
        if (target.hasParent) {
            utils_2.remove(source);
            utils_2.insert(source, target._parent);
            return true;
        }
        return false;
    }
    moveToBefore(source, target) {
        this._canMoveToSibling(source, target) && utils_2.setBefore(source, target);
        return this;
    }
    moveToAfter(source, target) {
        this._canMoveToSibling(source, target) && utils_2.setAfter(source, target);
        return this;
    }
    moveToChildrenFirst(source, target) {
        this._canMoveRemove(source) && this.addToChildrenFirst(source, target);
        return this;
    }
    moveToChildrenLast(source, target) {
        this._canMoveRemove(source) && this.addToChildrenLast(source, target);
        return this;
    }
    moveToReplace(source, target) {
        if (this._canMoveRemove(source)) {
            utils_2.setAfter(source, target);
            utils_2.remove(target);
        }
        return this;
    }
    //#endregion
    //#region delete node 
    separate(target) {
        utils_2.remove(target);
        return this;
    }
    remove(target) {
        utils_2.destroy(target);
        return this;
    }
    //#endregion
    //#region utils 
    isAncestor(source, ancestor) {
        let has = false;
        utils_2.bubble(source, (current) => {
            has = current === ancestor;
            return !has;
        });
        return has;
    }
    map(source, callback) {
        let mapNode = source;
        if (types_1.isFunction(source))
            mapNode = this._root;
        return utils_2.loopDFSTail(mapNode, (current, childrenResults) => callback(current, childrenResults));
    }
    bubble(source, fn) {
        utils_2.bubble(source, fn);
        return this;
    }
    import(source, { sourceNode = this._root, format, childrenField = 'children', useTreeForamt = true, coverRoot } = {}) {
        const sourceIsArray = Array.isArray(source);
        let result = utils_1.loopDFSTail(sourceIsArray ? { [childrenField]: source } : source, childrenField, (current, childrenResults) => {
            let tree = new LinkedTreeNode(useTreeForamt
                ? current.data
                : format
                    ? format(current.data)
                    : omit_1.default(current, childrenField));
            childrenResults.forEach(res => this.addToChildrenLast(res.tree, tree));
            return { tree };
        });
        result.tree.children.forEach((v) => this.addToChildrenLast(v, sourceNode));
        !sourceIsArray && coverRoot && (sourceNode.data = result.tree.data);
        return this;
    }
    export({ sourceNode = this._root, format, childrenField = 'children', useTreeForamt = true } = {}) {
        let result = utils_2.loopDFSTail(sourceNode, (current, childrenResults) => {
            let fmt = format
                ? format(current.data)
                : current.data;
            let children = childrenResults.map(res => res.fmtValue);
            let fmtValue = useTreeForamt
                ? { data: fmt, children }
                : Object.assign(Object.assign({}, fmt), { [childrenField]: children });
            return { fmtValue };
        });
        return result.fmtValue;
    }
    dispose() {
        utils_2.loopDFSTail(this._root, utils_2.destroy);
        return this;
    }
}
exports.LinkedTree = LinkedTree;
